/*****************************************************************************/
/*                   Reflections Plugin Support Functions                    */
/*****************************************************************************/

/* title:    list functions */
/* author:   Reinhard Grams  */
/* version:  1.0, 18.9.97 */
/* updated:  10.3.98 by RG */
/* updated:  21.11.99 by RG */

/*****************************************************************************/

#include <Lists_d.h>
#include <Lists_f.h>

/*****************************************************************************/
/* zero the list struct. *list is assumed to be valid */

   void
RPL_list_new
   (BasList *list)
{
   list->first=list->last=NULL;
   list->nr_of_nodes=0;
   list->index=NULL;
}

/*****************************************************************************/
/* add a node at the head of a list. *list is assumed to be valid */
   
   void
RPL_list_add_head
   (BasList *list,BasNode *node)
{
   node->pred=NULL;

   if (node->succ=list->first) {
      (node->succ)->pred=list->first=node;
   }
   else {
      list->first=list->last=node;
   }
   list->nr_of_nodes++;
}

/*****************************************************************************/
/* add a node at the tail of a list. *list is assumed to be valid */

   void
RPL_list_add_tail
   (BasList *list,BasNode *node)
{
   node->succ=NULL;

   if (node->pred=list->last) {
      (node->pred)->succ=list->last=node;
   }
   else {
      list->first=list->last=node;
   }
   list->nr_of_nodes++;
}

/*****************************************************************************/
/* add a node before a given node (position). ptrs are assumed to be valid */
/* 21.11.99 removed a bug and renamed the function, it formerly inserted a */
/* node after the given position */

   void
RPL_list_add_node_before
   (BasList *list,BasNode *position,BasNode *node)
{
   /* if list is empty, position is NULL or at start of list */
   if (list->first==NULL || position==NULL || position==list->first) {
      RPL_list_add_head(list,node);
   }
   else {
      node->succ=position;
      if (node->pred=position->pred) {
         (node->pred)->succ=node;
      }
      position->pred=node;
      list->nr_of_nodes++;
   }
}

/*****************************************************************************/
/* add a node after a given node (position). ptrs are assumed to be valid */
/* 22.11.99 added this function, which is a debugged version of formerly 
/* RPL_list_add_node, which was not correct */

   void
RPL_list_add_node_after
   (BasList *list,BasNode *position,BasNode *node)
{
   /* if list is empty, position is NULL or at end of list */
   if (list->first==NULL || position==NULL || position==list->last) {
      RPL_list_add_tail(list,node);
   }
   else {
      node->pred=position;
      if (node->succ=position->succ) {
         (node->succ)->pred=node;
      }
      position->succ=node;
      list->nr_of_nodes++;
   }
}

/*****************************************************************************/
/* remove the first node of a list. *list is assumed to be valid */

   void
RPL_list_remove_head
   (BasList *list)
{
   BasNode *first,*succ;

   if (first=list->first) {
      if (succ=first->succ) {
         succ->pred=NULL;
         list->first=succ;
      }
      else {
         list->first=list->last=NULL;
      }
      list->nr_of_nodes--;
   }
}

/*****************************************************************************/
/* remove the last node of a list. *list is assumed to be valid */

   void
RPL_list_remove_tail
   (BasList *list)
{
   BasNode *last,*pred;

   if (last=list->last) {
      if (pred=last->pred) {
         pred->succ=NULL;
         list->last=pred;
      }
      else {
         list->first=list->last=NULL;
      }
      list->nr_of_nodes--;
   }
}

/*****************************************************************************/
/* remove a node of a list. *list and *node are assumed to be valid */

   void
RPL_list_remove_node
   (BasList *list,BasNode *node)
{
   if (node==list->first) {
      RPL_list_remove_head(list);
   }
   else {
      if (node==list->last) {
         RPL_list_remove_tail(list);
      }
      else {
         (node->pred)->succ=node->succ;
         (node->succ)->pred=node->pred;
         list->nr_of_nodes--;
      }
   }
}

/*****************************************************************************/
/* build an index array of ptrs to the nodes. *list is assumed to be valid */

   BasNode **
RPL_list_create_index
   (BasList *list)
{
   BasNode *node,**index;

   if (list->nr_of_nodes>0) {
      if (list->index=index=RPI_alloc_mem(HND,
            list->nr_of_nodes*sizeof(BasNode *))) {
         node=list->first;
         while (node) {
            *index++=node; 
            node=node->succ;
         }
      }
   }
   return(list->index);
}

/*****************************************************************************/
/* remove an index array. *list is assumed to be valid */

   void
RPL_list_remove_index
   (BasList *list)
{
   if (list->index) {
      RPI_free_mem(HND,list->index);
      list->index=NULL;
   }
}

/*****************************************************************************/
/* revmove all list nodes and free node memory. *list is assumed to be valid */

   void
RPL_list_clear
   (BasList *list)
{
   BasNode *pred,*node;

   pred=list->last;
   while (node=pred) {
      pred=node->pred;
      RPI_free_mem(HND,node);
   }
   RPL_list_remove_index(list);
   RPL_list_new(list);
}

/*****************************************************************************/
/* build an index array of ptrs to the nodes. *list is assumed to be valid */

   BasNode **
RPL_list_create_heap_index
   (BasList *list,void *heap)
{
   BasNode *node,**index;

   if (list->nr_of_nodes>0) {
      if (list->index=index=RPI_heap_alloc_mem(HND,heap,
            list->nr_of_nodes*sizeof(BasNode *))) {
         node=list->first;
         while (node) {
            *index++=node; 
            node=node->succ;
         }
      }
   }
   return(list->index);
}

/*****************************************************************************/
/* remove an index array. *list is assumed to be valid */

   void
RPL_list_remove_heap_index
   (BasList *list,void *heap)
{
   if (list->index) {
      RPI_heap_free_mem(HND,heap,list->index);
      list->index=NULL;   
   }
}

/*****************************************************************************/
/* revmove all list nodes and free node memory. *list is assumed to be valid */

   void
RPL_list_heap_clear
   (BasList *list,void *heap)
{
   BasNode *pred,*node;

   pred=list->last;
   while (node=pred) {
      pred=node->pred;
      RPI_heap_free_mem(HND,heap,node);
   }
   RPL_list_remove_heap_index(list,heap);
   RPL_list_new(list);
}

/*****************************************************************************/
/* get a node by its position in list. *list is assumed to be valid */
/* index is assumed to be valid (list not altered after create_index call) */

   BasNode *
RPL_list_get_node
   (BasList *list,long number)
{
   BasNode *node;
   long current;

   if (number>=0 && number<list->nr_of_nodes) {

      if (list->index) return(list->index[number]);
         
      if (number<=(list->nr_of_nodes>>1)) {      
         current=0;
         node=list->first;
         while (node) {
            if (number==current++) return(node);
            node=node->succ;
         }
      }
      else {
         current=list->nr_of_nodes;
         node=list->last;
         while (node) {
            if (number==--current) return(node);
            node=node->pred;
         }
      }
   }
   return(NULL);
}

/*****************************************************************************/
/* dump the list and its nodes */

#ifdef _LISTS_DEBUG_

   void
RPL_list_dump_nodes
   (BasList *list)
{
   BasNode *node=list->first;
   char mess[128];

   sprintf(mess,"first %8lx last %8lx",list->first,list->last);
   RPI_message_win(HND,mess);

   while (node) {

      sprintf(mess,"node=%8lx succ=%8lx pred=%8lx",node,node->succ,node->pred);
      RPI_message_win(HND,mess);

      node=node->succ;
   }
}

#endif

/*****************************************************************************/

